from __future__ import annotations

import asyncio
import os
from itertools import chain
from typing import TYPE_CHECKING, Any, Optional

from pydantic import Field

from iambic.core.context import ctx
from iambic.core.iambic_enum import Command, IambicManaged
from iambic.core.logger import log
from iambic.core.models import (
    AccountChangeDetails,
    BaseModel,
    BaseTemplate,
    ExpiryModel,
    ProposedChange,
    ProposedChangeType,
    TemplateChangeDetails,
)
from iambic.core.utils import NoqSemaphore
from iambic.plugins.v0_1_0.okta.models import User, UserStatus
from iambic.plugins.v0_1_0.okta.user.utils import (
    create_user,
    get_user,
    maybe_deprovision_user,
    update_user_profile,
    update_user_status,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.okta.iambic_plugin import OktaConfig, OktaOrganization

OKTA_USER_TEMPLATE_TYPE = "NOQ::Okta::User"
OKTA_GET_USER_SEMAPHORE = NoqSemaphore(get_user, 10)


class Assignment(BaseModel):
    role: str
    resource_set: str


class UserProperties(BaseModel):
    username: str = Field(..., description="Username of the user")
    user_id: str = Field("", description="Unique User ID for the user")
    status: UserStatus = Field(UserStatus.active, description="Status of the user")
    profile: dict[str, Any]
    extra: Optional[dict[str, Any]] = Field(
        None, description=("Extra attributes to store for the user")
    )

    @classmethod
    def iambic_specific_knowledge(cls) -> set[str]:
        return {"extra", "metadata_commented_dict"}

    @property
    def resource_type(self) -> str:
        return "okta:user"

    @property
    def resource_id(self) -> str:
        return self.user_id


class OktaUserTemplate(BaseTemplate, ExpiryModel):
    template_type: str = "NOQ::Okta::User"
    properties: UserProperties
    force_delete: bool = Field(
        False,
        description=(
            "If `self.deleted` is true, the user will be force deleted from Okta. "
        ),
    )
    idp_name: str = Field(
        ...,
        description="Name of the identity provider that's associated with the user",
    )

    async def apply(
        self,
        config: OktaConfig,
    ) -> TemplateChangeDetails:
        tasks = []
        template_changes = TemplateChangeDetails(
            resource_id=self.properties.user_id,
            resource_type=self.template_type,
            template_path=self.file_path,
        )
        log_params = dict(
            resource_type=self.resource_type,
            resource_name=self.properties.username,
        )

        if self.iambic_managed == IambicManaged.IMPORT_ONLY:
            log_str = "Resource is marked as import only."
            log.info(log_str, **log_params)
            template_changes.proposed_changes = []
            return template_changes

        for okta_organization in config.organizations:
            if self.idp_name != okta_organization.idp_name:
                continue
            if ctx.execute:
                log_str = "Applying changes to resource."
            else:
                log_str = "Detecting changes for resource."
            log.info(log_str, idp_name=okta_organization.idp_name, **log_params)
            tasks.append(self._apply_to_account(okta_organization))

        account_changes = await asyncio.gather(*tasks)
        template_changes.proposed_changes = [
            account_change
            for account_change in account_changes
            if any(account_change.proposed_changes)
        ]

        proposed_changes = [x for x in account_changes if x.proposed_changes]

        if proposed_changes and ctx.execute:
            log.info(
                "Successfully applied all or some resource changes to all Okta organizations. Any unapplied resources will have an accompanying error message.",
                **log_params,
            )
        elif proposed_changes:
            log.info(
                "Successfully detected all or some required resource changes on all Okta organizations. Any unapplied resources will have an accompanying error message.",
                **log_params,
            )
        else:
            log.debug("No changes detected for resource on any account.", **log_params)

        return template_changes

    @property
    def resource_id(self) -> str:
        return self.properties.user_id

    @property
    def resource_type(self) -> str:
        return "okta:user"

    def set_default_file_path(self, repo_dir: str):
        file_name = f"{self.properties.username}.yaml"
        self.file_path = os.path.expanduser(
            os.path.join(repo_dir, f"resources/okta/user/{self.idp_name}/{file_name}")
        )

    def apply_resource_dict(
        self,
        okta_organization: OktaOrganization,
    ):
        return {
            "username": self.properties.username,
            "status": self.properties.status,
            "profile": self.properties.profile,
        }

    async def _apply_to_account(
        self,
        okta_organization: OktaOrganization,
    ) -> AccountChangeDetails:
        await self.remove_expired_resources()
        proposed_user = self.apply_resource_dict(okta_organization)
        change_details = AccountChangeDetails(
            account=self.idp_name,
            resource_id=self.properties.username,
            resource_type=self.properties.resource_type,
            new_value=proposed_user,
            proposed_changes=[],
        )

        log_params = dict(
            resource_type=self.properties.resource_type,
            resource_id=self.properties.username,
            organization=str(self.idp_name),
        )

        current_user_task = await OKTA_GET_USER_SEMAPHORE.process(
            [
                {
                    "username": self.properties.username,
                    "user_id": self.properties.user_id,
                    "okta_organization": okta_organization,
                }
            ]
        )
        current_user: Optional[User] = current_user_task[0]
        if current_user:
            change_details.current_value = current_user

            if ctx.command == Command.CONFIG_DISCOVERY:
                # Don't overwrite a resource during config discovery
                change_details.new_value = {}
                return change_details

        user_exists = bool(current_user)
        tasks = []
        if self.deleted:
            self.properties.status = UserStatus.deprovisioned
            proposed_user["status"] = "deprovisioned"

        if not user_exists and not self.deleted:
            change_details.extend_changes(
                [
                    ProposedChange(
                        change_type=ProposedChangeType.CREATE,
                        resource_id=self.properties.username,
                        resource_type=self.properties.resource_type,
                    )
                ]
            )
            log_str = "New resource found in code."
            if not ctx.execute:
                log.info(log_str, **log_params)
                # Exit now because apply functions won't work if resource doesn't exist
                return change_details

            log_str = f"{log_str} Creating resource..."
            log.info(log_str, **log_params)

            current_user: User = await create_user(
                self,
                okta_organization=okta_organization,
            )
            if current_user:
                change_details.current_value = current_user
                self.properties.user_id = current_user.user_id
                self.write()
        if (
            current_user
            and self.deleted
            and current_user.status == UserStatus.deprovisioned
            and proposed_user.get("status") == "deprovisioned"
            and not self.force_delete
        ):
            log.info(
                "User is already deprovisioned. Please delete the user in Okta.",
                **log_params,
            )
            return change_details

        if current_user and not self.deleted:
            tasks.append(
                update_user_status(
                    current_user,
                    self.properties.status.value,
                    okta_organization,
                    log_params,
                ),
            )

        tasks.extend(
            [
                update_user_profile(
                    self,
                    current_user,
                    self.properties.profile,
                    okta_organization,
                    log_params,
                ),
                maybe_deprovision_user(
                    bool(self.deleted),
                    current_user,
                    okta_organization,
                    log_params,
                ),
            ]
        )

        changes_made = await asyncio.gather(*tasks)
        if any(changes_made):
            change_details.extend_changes(list(chain.from_iterable(changes_made)))

        if ctx.execute:
            log.debug(
                "Successfully finished execution for resource",
                changes_made=bool(change_details.proposed_changes),
                **log_params,
            )
            self.write()
        else:
            log.debug(
                "Successfully finished scanning for drift for resource",
                requires_changes=bool(change_details.proposed_changes),
                **log_params,
            )
        return change_details



from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, List, Optional

from aiohttp import ClientResponseError

from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import GlobalRetryController, snake_to_camelback
from iambic.plugins.v0_1_0.azure_ad.group.models import (
    GroupTemplateProperties,
    Member,
    MemberDataType,
)
from iambic.plugins.v0_1_0.azure_ad.models import AzureADOrganization
from iambic.plugins.v0_1_0.azure_ad.user.utils import get_user

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.azure_ad.user.models import UserTemplateProperties


async def list_group_members(
    azure_ad_organization: AzureADOrganization,
    group: GroupTemplateProperties,
) -> GroupTemplateProperties:
    """
    List the members of a group in Azure AD.

    Args:
    - azure_ad_organization: An instance of the AzureADOrganization class, which provides access to the Microsoft Graph API.
    - group: An instance of the `GroupTemplateProperties` class, representing the group whose members we want to list.

    Returns:
    - The same instance of the `GroupTemplateProperties` class, with the `members` attribute populated with a list of `User` instances,
    representing the members of the group.
    """
    async with GlobalRetryController(
        fn_identifier="azure_ad.list_group_members"
    ) as retry_controller:
        fn = functools.partial(
            azure_ad_organization.list, f"groups/{group.group_id}/members"
        )
        members = await retry_controller(fn)

    if not members:
        return group

    user_members = list(
        await asyncio.gather(
            *[
                get_user(
                    azure_ad_organization,
                    user_id=member.get("id"),
                    allow_template_ref=True,
                )
                for member in members
                if member.get("@odata.type").endswith("user")
            ],
            return_exceptions=True,
        )
    )
    group.members = [
        Member(id=member.user_id, name=member.username, data_type=MemberDataType.USER)
        for member in user_members
    ]

    if len(user_members) != len(members):
        for member in members:
            if not member.get("@odata.type").endswith("group"):
                continue

            if (mail := member.get("mail")) and "@" in mail:
                name = mail
            else:
                name = member.get("displayName")

            group.members.append(
                Member(id=member["id"], name=name, data_type=MemberDataType.GROUP)
            )

    return group


async def list_groups(
    azure_ad_organization: AzureADOrganization, include_members: bool = True, **kwargs
) -> List[GroupTemplateProperties]:
    from iambic.plugins.v0_1_0.azure_ad.group.models import GroupTemplateProperties

    async with GlobalRetryController(
        fn_identifier="azure_ad.list_groups"
    ) as retry_controller:
        fn = functools.partial(azure_ad_organization.list, "groups", **kwargs)
        groups = await retry_controller(fn)
        groups = [GroupTemplateProperties.from_azure_response(g) for g in groups]

    if not include_members:
        return groups

    tasks = []
    for group in groups:
        tasks.append(list_group_members(azure_ad_organization, group))
    return list(await asyncio.gather(*tasks))


async def get_group(
    azure_ad_organization: AzureADOrganization,
    group_id: Optional[str] = None,
    group_name: Optional[str] = None,
) -> Optional[GroupTemplateProperties]:
    """
    Get a group from Azure AD using the Microsoft Graph API.

    Args:
    - azure_ad_organization: An instance of the AzureADOrganization class, which provides access to the Microsoft Graph API.
    - group_id: The ID of the group to get.
    - group_name: The name of the group to get.

    Returns:
    - An instance of the `GroupTemplateProperties` class, representing the retrieved group. If an error occurs, returns None.
    """
    assert group_id or group_name, "Must provide either group_id or group_name"

    if group_id:
        async with GlobalRetryController(
            fn_identifier="azure_ad.get_group"
        ) as retry_controller:
            fn = functools.partial(azure_ad_organization.get, f"groups/{group_id}")
            if group := await retry_controller(fn):
                group = GroupTemplateProperties.from_azure_response(group)
                return await list_group_members(azure_ad_organization, group)
            elif not group_name:
                raise Exception(f"Group not found with id {group_id}")

    # Try to get group by name
    groups = await list_groups(
        azure_ad_organization,
        include_members=True,
        params={"$filter": f"displayName eq '{group_name}'"},
    )
    if groups:
        return groups[0]

    raise Exception(f"Group not found with displayName {group_name}")


async def create_group(
    azure_ad_organization: AzureADOrganization,
    group_name: str,
    description: str,
    mail_enabled: bool,
    mail_nickname: Optional[str],
    security_enabled: bool,
    group_types: list[str],
) -> Optional[GroupTemplateProperties]:
    """
    Create a new group in Azure AD.
    Args:
    - azure_ad_organization (AzureADOrganization): The Azure AD organization to update the group in.
    - group_name: The name of the group to create.
    - idp_name: The IDP name for the group.
    - description: The description for the group.

    Returns:
    - An instance of the `GroupTemplateProperties` class, representing the created group. If an error occurs, returns None.
    """

    group = await azure_ad_organization.post(
        "groups",
        json={
            "displayName": group_name,
            "mailEnabled": mail_enabled,
            "mailNickname": mail_nickname,
            "securityEnabled": security_enabled,
            "description": description,
            "groupTypes": group_types,
        },
    )
    return GroupTemplateProperties.from_azure_response(group)


async def update_group_attributes(
    azure_ad_organization: AzureADOrganization,
    template_group: GroupTemplateProperties,
    cloud_group: GroupTemplateProperties,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Update the name of a group in Azure AD.
        Args:
        azure_ad_organization (AzureADOrganization): The Azure AD organization to update the group in.
        template_user (GroupTemplateProperties): The template representation of the group.
        cloud_user (GroupTemplateProperties): The current representation of the group in the cloud.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list[ProposedChange] = []
    patch_request = {}

    for attr, value in cloud_group.dict(
        exclude_none=False, exclude={"group_id", "members", "mail"}
    ).items():
        if (template_value := getattr(template_group, attr)) != value:
            response.append(
                ProposedChange(
                    change_type=ProposedChangeType.UPDATE,
                    resource_id=template_group.resource_id,
                    resource_type=template_group.resource_type,
                    attribute=attr,
                    current_value=value,
                    new_value=template_value,
                )
            )

            attr = "name" if attr == "display_name" else snake_to_camelback(attr)
            patch_request[attr] = template_value

    if ctx.execute and patch_request:
        try:
            await azure_ad_organization.patch(
                f"groups/{template_group.group_id}",
                json=patch_request,
            )
        except ClientResponseError as err:
            log.exception(
                "Failed to update group in Azure AD",
                **log_params,
            )
            response[0].exceptions_seen = [str(err)]

    return response


def parse_group_member_response(
    members: list[Member],
    proposed_change: ProposedChange,
    responses: list,
    log_params: dict,
) -> ProposedChange:
    for i, response in enumerate(responses):
        if isinstance(response, ClientResponseError):
            member = members[i]
            log.exception(
                "Failed to update group members in Azure AD",
                err=str(response),
                member=member.dict(),
                **log_params,
            )
            proposed_change.exceptions_seen.append(
                f"Failed to add or remove member. "
                f"Member: {member.name} ({member.data_type}). "
                f"Response: {response.status} - {response.message}."
            )
    return proposed_change


async def update_group_members(
    azure_ad_organization: AzureADOrganization,
    cloud_group: GroupTemplateProperties,
    template_members: List[Member],
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Update the members of a group in Azure AD.

    Args:
        azure_ad_organization (AzureADOrganization): The Azure AD organization to update the group in.
        cloud_group (GroupTemplateProperties): The group to update the members of.
        template_members (List[Member]): The new members to add to the group.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response = []

    current_member_ids = [member.id for member in cloud_group.members]
    desired_member_ids = [member.id for member in template_members]

    members_to_remove = [
        member for member in cloud_group.members if member.id not in desired_member_ids
    ]
    members_to_add = [
        member for member in template_members if member.id not in current_member_ids
    ]

    if members_to_remove:
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_id=cloud_group.resource_id,
                resource_type=cloud_group.resource_type,
                attribute="members",
                change_summary={
                    "MembersToRemove": [member.dict() for member in members_to_remove]
                },
                current_value=[member.name for member in cloud_group.members],
                new_value=[member.name for member in members_to_remove],
            )
        )

    if members_to_add:
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_id=cloud_group.resource_id,
                resource_type=cloud_group.resource_type,
                attribute="members",
                change_summary={
                    "MembersToAdd": [member.dict() for member in members_to_add]
                },
                current_value=[member.name for member in cloud_group.members],
                new_value=[member.name for member in members_to_add],
            )
        )

    if ctx.execute:
        tasks = []
        if members_to_remove:
            for member in members_to_remove:
                tasks.append(
                    azure_ad_organization.delete(
                        f"groups/{cloud_group.group_id}/members/{member.id}/$ref"
                    )
                )
        if members_to_add:
            for member in members_to_add:
                tasks.append(
                    azure_ad_organization.post(
                        f"groups/{cloud_group.group_id}/members/$ref",
                        json={
                            "@odata.id": f"https://graph.windows.net/v1.0/directoryObjects/{member.id}"
                        },
                    )
                )
        responses = await asyncio.gather(*tasks, return_exceptions=True)

        if members_to_remove:
            response[0] = parse_group_member_response(
                members=members_to_remove,
                proposed_change=response[0],
                responses=list(responses[: len(members_to_remove)]),
                log_params=log_params,
            )

        if members_to_add:
            response_elem = 0 if not members_to_remove else 1
            response[response_elem] = parse_group_member_response(
                members=members_to_add,
                proposed_change=response[response_elem],
                responses=list(responses[len(members_to_remove) :]),
                log_params=log_params,
            )

    return response


async def delete_group(
    azure_ad_organization: AzureADOrganization,
    group: GroupTemplateProperties,
    log_params: dict[str, str],
) -> List[ProposedChange]:
    """
    Delete a group in Azure AD.
    Args:
        azure_ad_organization (AzureADOrganization): The Azure AD organization to delete the group from.
        group (GroupTemplateProperties): The group to delete.
        log_params (dict): Logging parameters.

    Returns:
        List[ProposedChange]: A list of proposed changes to be applied.
    """
    response: list[ProposedChange] = [
        ProposedChange(
            change_type=ProposedChangeType.DELETE,
            resource_id=group.resource_id,
            resource_type=group.resource_type,
            attribute="group",
            change_summary={"group": group.name},
            current_value=group.name,
            new_value=None,
        )
    ]

    if ctx.execute:
        try:
            await azure_ad_organization.delete(f"groups/{group.group_id}")
        except ClientResponseError as err:
            log.exception(
                "Failed to delete group in Azure AD",
                **log_params,
            )
            response[0].exceptions_seen = [str(err)]

    return response


async def resolve_member_ids(
    azure_ad_organization: AzureADOrganization, members: List[Member]
) -> List[Member]:
    from iambic.plugins.v0_1_0.azure_ad.group.utils import get_group
    from iambic.plugins.v0_1_0.azure_ad.user.utils import get_user

    for member in members:
        if member.id:
            continue
        if member.data_type == MemberDataType.USER:
            user_details: Optional[UserTemplateProperties] = await get_user(
                azure_ad_organization, username=member.name
            )
            if not user_details:
                continue
            member.id = user_details.user_id
        elif member.data_type == MemberDataType.GROUP:
            group_details: Optional[GroupTemplateProperties] = await get_group(
                azure_ad_organization, group_name=member.name
            )
            if not group_details:
                continue
            member.id = group_details.group_id
    return members



from __future__ import annotations

import asyncio
from itertools import chain

from botocore.exceptions import ClientError
from deepdiff import DeepDiff

from iambic.core import noq_json as json
from iambic.core.context import ctx
from iambic.core.logger import log
from iambic.core.models import ProposedChange, ProposedChangeType
from iambic.core.utils import aio_wrapper, async_batch_processor, plugin_apply_wrapper
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.utils import boto_crud_call, legacy_paginated_search


async def get_permission_set_details(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
) -> dict:
    try:
        return (
            await boto_crud_call(
                identity_center_client.describe_permission_set,
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
            )
        ).get("PermissionSet", {})
    except identity_center_client.exceptions.ResourceNotFoundException:
        return {}
    except ClientError as e:
        if e.response["Error"]["Code"] == "ResourceNotFound":
            return {}
        else:
            raise


async def generate_permission_set_map(aws_accounts: list[AWSAccount], templates: list):
    """Generates the map of permission sets for AWS accounts that are referenced in at least 1 template

    :param aws_accounts:
    :param templates:
    :return:
    """
    from iambic.plugins.v0_1_0.aws.identity_center.permission_set.models import (
        AWS_IDENTITY_CENTER_PERMISSION_SET_TEMPLATE_TYPE,
    )

    org_include_set = set()
    include_all = False
    accounts_to_set_identity_center = []

    for template in templates:
        if template.template_type == AWS_IDENTITY_CENTER_PERMISSION_SET_TEMPLATE_TYPE:
            for org in template.included_orgs:
                if org == "*":
                    include_all = True
                    break
                org_include_set.add(org)

        if include_all:
            break

    for aws_account in aws_accounts:
        if (
            include_all or aws_account.org_id in org_include_set
        ) and aws_account.identity_center_details:
            accounts_to_set_identity_center.append(aws_account)

    await asyncio.gather(
        *[
            account.set_identity_center_details()
            for account in accounts_to_set_identity_center
        ]
    )


async def get_permission_set_users_and_groups(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    user_map: dict,
    group_map: dict,
) -> dict:
    response = {
        "user": {k: dict(accounts=[], **v) for k, v in user_map.items()},
        "group": {k: dict(accounts=[], **v) for k, v in group_map.items()},
    }

    query_params = dict(InstanceArn=instance_arn, PermissionSetArn=permission_set_arn)
    instance_accounts = await legacy_paginated_search(
        identity_center_client.list_accounts_for_provisioned_permission_set,
        response_key="AccountIds",
        **query_params,
    )
    all_account_assignments = await async_batch_processor(
        [
            legacy_paginated_search(
                identity_center_client.list_account_assignments,
                response_key="AccountAssignments",
                AccountId=account_id,
                **query_params,
            )
            for account_id in instance_accounts
        ],
        15,
        1,
        return_exceptions=True,
    )

    for account_assignments in all_account_assignments:
        for aa in account_assignments:
            response[aa["PrincipalType"].lower()][aa["PrincipalId"]]["accounts"].append(
                aa["AccountId"]
            )

    response["user"] = {k: v for k, v in response["user"].items() if v["accounts"]}
    response["group"] = {k: v for k, v in response["group"].items() if v["accounts"]}
    return response


async def get_permission_set_users_and_groups_as_access_rules(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    user_map: dict,
    group_map: dict,
    org_account_map: dict,
) -> list[dict]:
    name_map = {"user": "UserName", "group": "DisplayName"}
    permission_set_users_and_groups = await get_permission_set_users_and_groups(
        identity_center_client, instance_arn, permission_set_arn, user_map, group_map
    )
    response = []

    for assignment_type in ["user", "group"]:
        resource_type = assignment_type.upper()
        for assignment_id, assignment_details in permission_set_users_and_groups[
            assignment_type
        ].items():
            for account_id in assignment_details["accounts"]:
                response.append(
                    {
                        "account_id": account_id,
                        "resource_id": assignment_id,
                        "resource_type": resource_type,
                        "resource_name": assignment_details.get(
                            name_map[assignment_type]
                        ),
                        "account_name": f"{account_id} ({org_account_map[account_id]})",
                    }
                )

    return response


async def enrich_permission_set_details(
    identity_center_client, instance_arn: str, permission_set_details: dict
):
    permission_set_arn = permission_set_details["PermissionSetArn"]
    query_params = dict(InstanceArn=instance_arn, PermissionSetArn=permission_set_arn)
    tasks = [
        boto_crud_call(
            identity_center_client.get_inline_policy_for_permission_set, **query_params
        ),
        boto_crud_call(
            identity_center_client.get_permissions_boundary_for_permission_set,
            **query_params,
        ),
        legacy_paginated_search(
            identity_center_client.list_customer_managed_policy_references_in_permission_set,
            response_key="CustomerManagedPolicyReferences",
            retain_key=True,
            **query_params,
        ),
        legacy_paginated_search(
            identity_center_client.list_managed_policies_in_permission_set,
            response_key="AttachedManagedPolicies",
            retain_key=True,
            **query_params,
        ),
        legacy_paginated_search(
            identity_center_client.list_tags_for_resource,
            response_key="Tags",
            retain_key=True,
            InstanceArn=instance_arn,
            ResourceArn=permission_set_arn,
        ),
    ]
    permission_set_responses = await asyncio.gather(*tasks, return_exceptions=True)
    for permission_set_response in permission_set_responses:
        if isinstance(permission_set_response, Exception):
            continue

        for k, v in permission_set_response.items():
            if k == "ResponseMetadata" or not v:
                continue
            permission_set_details[k] = v
    return permission_set_details


async def apply_permission_set_aws_managed_policies(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    template_policy_arns: list[str],
    existing_policy_arns: list[str],
    log_params: dict,
) -> list[ProposedChange]:
    tasks = []
    response = []

    # Create new managed policies
    new_managed_policies = [
        policy_arn
        for policy_arn in template_policy_arns
        if policy_arn not in existing_policy_arns
    ]
    if new_managed_policies:
        log_str = "New AWS managed policies discovered."
        if ctx.execute:
            log_str = f"{log_str} Attaching AWS managed policies..."

        for policy_arn in new_managed_policies:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.ATTACH,
                    resource_type="aws:policy_document",
                    resource_id=policy_arn,
                    attribute="managed_policies",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                apply_awaitable = boto_crud_call(
                    identity_center_client.attach_managed_policy_to_permission_set,
                    InstanceArn=instance_arn,
                    PermissionSetArn=permission_set_arn,
                    ManagedPolicyArn=policy_arn,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, managed_policies=new_managed_policies, **log_params)

    # Delete existing managed policies not in template
    existing_managed_policies = [
        policy_arn
        for policy_arn in existing_policy_arns
        if policy_arn not in template_policy_arns
    ]
    if existing_managed_policies:
        log_str = "Stale AWS managed policies discovered."
        if ctx.execute:
            log_str = f"{log_str} Detaching AWS managed policies..."

        for policy_arn in existing_managed_policies:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.DETACH,
                    resource_type="aws:policy_document",
                    resource_id=policy_arn,
                    attribute="managed_policies",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                apply_awaitable = boto_crud_call(
                    identity_center_client.detach_managed_policy_from_permission_set,
                    InstanceArn=instance_arn,
                    PermissionSetArn=permission_set_arn,
                    ManagedPolicyArn=policy_arn,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, managed_policies=existing_managed_policies, **log_params)

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def apply_permission_set_customer_managed_policies(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    template_policies: list[dict],
    existing_policies: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    tasks = []
    response = []

    log.debug(
        "Customer Managed Policies",
        template_policies=template_policies,
        existing_policies=existing_policies,
    )

    template_policy_map = {
        f"{policy['Path']}{policy['Name']}": policy for policy in template_policies
    }
    existing_policy_map = {
        f"{policy['Path']}{policy['Name']}": policy for policy in existing_policies
    }

    # Create new managed policies
    new_customer_managed_policy_references = [
        policy
        for policy_path, policy in template_policy_map.items()
        if not existing_policy_map.get(policy_path)
    ]
    if new_customer_managed_policy_references:
        log_str = "New customer managed policies discovered."
        if ctx.execute:
            log_str = f"{log_str} Attaching customer managed policies..."

        for policy in new_customer_managed_policy_references:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.ATTACH,
                    resource_type="aws:policy_document",
                    resource_id=f"{policy['Path']}{policy['Name']}",
                    attribute="customer_managed_policies",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                apply_awaitable = boto_crud_call(
                    identity_center_client.attach_customer_managed_policy_reference_to_permission_set,
                    InstanceArn=instance_arn,
                    PermissionSetArn=permission_set_arn,
                    CustomerManagedPolicyReference=policy,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(
            log_str,
            customer_managed_policy_refs=new_customer_managed_policy_references,
            **log_params,
        )

    # Delete existing managed policies not in template
    existing_customer_managed_policy_references = [
        policy
        for policy_path, policy in existing_policy_map.items()
        if not template_policy_map.get(policy_path)
    ]
    if existing_customer_managed_policy_references:
        log_str = "Stale customer managed policies discovered."
        if ctx.execute:
            log_str = f"{log_str} Detaching customer managed policies..."

        for policy in existing_customer_managed_policy_references:
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.DETACH,
                    resource_type="aws:policy_document",
                    resource_id=f"{policy['Path']}{policy['Name']}",
                    attribute="customer_managed_policies",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                apply_awaitable = boto_crud_call(
                    identity_center_client.detach_customer_managed_policy_reference_from_permission_set,
                    retryable_errors=["ConflictException"],
                    InstanceArn=instance_arn,
                    PermissionSetArn=permission_set_arn,
                    CustomerManagedPolicyReference=policy,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(
            log_str,
            customer_managed_policy_refs=existing_customer_managed_policy_references,
            **log_params,
        )

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def create_account_assignment(
    identity_center_client,
    account_id: str,
    instance_arn: str,
    permission_set_arn: str,
    resource_type: str,
    resource_id: str,
    resource_name: str,
    log_params: dict,
):
    status = await boto_crud_call(
        identity_center_client.create_account_assignment,
        InstanceArn=instance_arn,
        TargetId=account_id,
        TargetType="AWS_ACCOUNT",
        PermissionSetArn=permission_set_arn,
        PrincipalType=resource_type,
        PrincipalId=resource_id,
    )
    request_id = status.get("AccountAssignmentCreationStatus", {}).get("RequestId")
    creation_status = status.get("AccountAssignmentCreationStatus", {})

    while creation_status.get("Status") == "IN_PROGRESS":
        await asyncio.sleep(0.5)
        status = await boto_crud_call(
            identity_center_client.describe_account_assignment_creation_status,
            InstanceArn=instance_arn,
            AccountAssignmentCreationRequestId=request_id,
        )
        creation_status = status.get("AccountAssignmentCreationStatus", {})

        if creation_status.get("Status") == "FAILED":
            log_params = {
                **log_params,
                "resource_type": f"aws:identity_center:account_assignment:{resource_type.lower()}",
            }
            log.error(
                "Unable to create account assignment.",
                reason=creation_status.get("FailureReason"),
                assigned_account_id=account_id,
                resource_name=resource_name,
                **log_params,
            )
            return


async def delete_account_assignment(
    identity_center_client,
    account_id: str,
    instance_arn: str,
    permission_set_arn: str,
    resource_type: str,
    resource_id: str,
    resource_name: str,
    log_params: dict,
):
    status = await boto_crud_call(
        identity_center_client.delete_account_assignment,
        InstanceArn=instance_arn,
        TargetId=account_id,
        TargetType="AWS_ACCOUNT",
        PermissionSetArn=permission_set_arn,
        PrincipalType=resource_type,
        PrincipalId=resource_id,
    )
    request_id = status.get("AccountAssignmentDeletionStatus", {}).get("RequestId")
    deletion_status = status.get("AccountAssignmentDeletionStatus", {})

    while deletion_status.get("Status") == "IN_PROGRESS":
        await asyncio.sleep(0.5)
        status = await boto_crud_call(
            identity_center_client.describe_account_assignment_deletion_status,
            InstanceArn=instance_arn,
            AccountAssignmentDeletionRequestId=request_id,
        )
        deletion_status = status.get("AccountAssignmentDeletionStatus", {})

        if deletion_status.get("Status") == "FAILED":
            log_params = {
                **log_params,
                "resource_type": f"aws:identity_center:account_assignment:{resource_type.lower()}",
            }
            log.error(
                "Unable to delete account assignment.",
                reason=deletion_status.get("FailureReason"),
                assigned_account_id=account_id,
                resource_name=resource_name,
                **log_params,
            )
            return


async def apply_account_assignments(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    template_assignments: list[dict],
    existing_assignments: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    """
    assignment objects = dict(
        account_id: str
        resource_id: str
        resource_name: str
        account_name: str
        resource_type: USER | GROUP
    )

    identity_center_client: boto3 client
    instance_arn: str
    permission_set_arn: str
    template_assignments: list[dict]. This is the desired list of assignments from the template
    existing_assignments: list[dict]. These are the existing assignments from AWS
    log_params: dict
    """
    tasks = []
    response = []
    template_assignment_map = {
        f"{ta['account_id']}:{ta['resource_type']}:{ta['resource_id']}": ta
        for ta in template_assignments
    }
    existing_assignment_map = {
        f"{ea['account_id']}:{ea['resource_type']}:{ea['resource_id']}": ea
        for ea in existing_assignments
    }

    for assignment_id, assignment in existing_assignment_map.items():
        if not template_assignment_map.get(assignment_id):
            log_str = "Stale assignments discovered."
            resource_type = (
                "arn:aws:iam::aws:user"
                if assignment["resource_type"] == "USER"
                else "arn:aws:iam::aws:group"
            )
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.DELETE,
                    account=assignment["account_name"],
                    resource_id=assignment["resource_name"],
                    resource_type=resource_type,
                    attribute="account_assignment",
                )
            ]
            response.extend(proposed_changes)
            if ctx.execute:
                log_str = f"{log_str} Removing account assignment..."
                apply_awaitable = delete_account_assignment(
                    identity_center_client,
                    account_id=assignment["account_id"],
                    instance_arn=instance_arn,
                    permission_set_arn=permission_set_arn,
                    resource_type=assignment["resource_type"],
                    resource_id=assignment["resource_id"],
                    resource_name=assignment["resource_name"],
                    log_params=log_params,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
            log.info(log_str, details=assignment, **log_params)

    for assignment_id, assignment in template_assignment_map.items():
        if not existing_assignment_map.get(assignment_id):
            resource_type = (
                "arn:aws:iam::aws:user"
                if assignment["resource_type"] == "USER"
                else "arn:aws:iam::aws:group"
            )
            proposed_changes = [
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    account=assignment["account_name"],
                    resource_id=assignment["resource_name"],
                    resource_type=resource_type,
                    attribute="account_assignment",
                )
            ]
            response.extend(proposed_changes)

            log_str = "New account assignments discovered."
            if ctx.execute:
                log_str = f"{log_str} Creating account assignment..."
                apply_awaitable = create_account_assignment(
                    identity_center_client,
                    account_id=assignment["account_id"],
                    instance_arn=instance_arn,
                    permission_set_arn=permission_set_arn,
                    resource_type=assignment["resource_type"],
                    resource_id=assignment["resource_id"],
                    resource_name=assignment["resource_name"],
                    log_params=log_params,
                )
                tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
            log.info(log_str, details=assignment, **log_params)

    if tasks:
        results: list[list[ProposedChange]] = await async_batch_processor(tasks, 10, 1)
        return list(chain.from_iterable(results))
    else:
        return response


async def apply_permission_set_inline_policy(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    template_inline_policy: str,
    existing_inline_policy: str,
    log_params: dict,
) -> list[ProposedChange]:
    response = []
    policy_drift = None

    if isinstance(template_inline_policy, str):
        template_inline_policy = json.loads(template_inline_policy)

    if existing_inline_policy:
        if isinstance(existing_inline_policy, str):
            existing_inline_policy = json.loads(existing_inline_policy)

        policy_drift = await aio_wrapper(
            DeepDiff,
            existing_inline_policy,
            template_inline_policy,
            report_repetition=True,
            ignore_order=True,
        )

    if template_inline_policy and (not existing_inline_policy or bool(policy_drift)):
        log_str = "Changes to the InlinePolicyDocument discovered."
        if policy_drift:
            response.append(
                ProposedChange(
                    change_type=ProposedChangeType.UPDATE,
                    attribute="inline_policy_document",
                    resource_id=permission_set_arn,
                    resource_type="aws:identity_center:permission_set",
                    change_summary=policy_drift,
                    current_value=existing_inline_policy,
                    new_value=template_inline_policy,
                )
            )
        else:
            response.append(
                ProposedChange(
                    change_type=ProposedChangeType.CREATE,
                    resource_type="aws:identity_center:permission_set",
                    resource_id=permission_set_arn,
                    attribute="inline_policy_document",
                    new_value=template_inline_policy,
                )
            )

        if ctx.execute:
            boto_action = "Creating" if not existing_inline_policy else "Updating"
            log_str = f"{log_str} {boto_action} InlinePolicyDocument..."
            apply_awaitable = boto_crud_call(
                identity_center_client.put_inline_policy_to_permission_set,
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
                InlinePolicy=json.dumps(template_inline_policy),
            )
            response = await plugin_apply_wrapper(apply_awaitable, response)

        log.debug(log_str, **log_params)
    elif not template_inline_policy and existing_inline_policy:
        log_str = "Stale InlinePolicyDocument."
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.DELETE,
                attribute="inline_policy",
                resource_type="aws:identity_center:permission_set",
                resource_id=permission_set_arn,
                current_value=existing_inline_policy,
            )
        )
        if ctx.execute:
            log_str = f"{log_str} Removing InlinePolicyDocument..."
            apply_awaitable = boto_crud_call(
                identity_center_client.delete_inline_policy_from_permission_set,
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
            )
            response = await plugin_apply_wrapper(apply_awaitable, response)

        log.debug(log_str, **log_params)

    return response


async def apply_permission_set_permission_boundary(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    template_permission_boundary: dict,
    existing_permission_boundary: dict,
    log_params: dict,
) -> list[ProposedChange]:
    response = []
    policy_drift = None

    if existing_permission_boundary:
        policy_drift = await aio_wrapper(
            DeepDiff,
            existing_permission_boundary,
            template_permission_boundary,
            report_repetition=True,
            ignore_order=True,
        )

    if template_permission_boundary and (
        not existing_permission_boundary or bool(policy_drift)
    ):
        log_str = "Changes to the PermissionsBoundary discovered."
        if policy_drift:
            response.append(
                ProposedChange(
                    change_type=ProposedChangeType.UPDATE,
                    attribute="permissions_boundary",
                    resource_type="aws:identity_center:permission_set",
                    resource_id=permission_set_arn,
                    change_summary=policy_drift,
                    current_value=existing_permission_boundary,
                    new_value=template_permission_boundary,
                )
            )
        else:
            response.append(
                ProposedChange(
                    change_type=ProposedChangeType.ATTACH,
                    resource_type="aws:identity_center:permission_set",
                    resource_id=permission_set_arn,
                    attribute="permissions_boundary",
                    new_value=template_permission_boundary,
                )
            )

        if ctx.execute:
            boto_action = "Creating" if not existing_permission_boundary else "Updating"
            log_str = f"{log_str} {boto_action} PermissionsBoundary..."
            apply_awaitable = boto_crud_call(
                identity_center_client.put_permissions_boundary_to_permission_set,
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
                PermissionsBoundary=template_permission_boundary,
            )
            response = await plugin_apply_wrapper(apply_awaitable, response)

        log.debug(log_str, **log_params)
    elif existing_permission_boundary and not template_permission_boundary:
        log_str = "Removing PermissionsBoundary discovered."
        response.append(
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_type="aws:identity_center:permission_set",
                resource_id=permission_set_arn,
                attribute="permissions_boundary",
                current_value=existing_permission_boundary,
            )
        )

        if ctx.execute:
            log_str = f"{log_str} Deleting PermissionsBoundary..."
            apply_awaitable = boto_crud_call(
                identity_center_client.delete_permissions_boundary_from_permission_set,
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
            )
            response = await plugin_apply_wrapper(apply_awaitable, response)

        log.debug(log_str, **log_params)

    return response


async def apply_permission_set_tags(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    template_tags: list[dict],
    existing_tags: list[dict],
    log_params: dict,
) -> list[ProposedChange]:
    existing_tag_map = {tag["Key"]: tag.get("Value") for tag in existing_tags}
    template_tag_map = {tag["Key"]: tag.get("Value") for tag in template_tags}
    tags_to_apply = [
        tag
        for tag in template_tags
        if tag.get("Value") != existing_tag_map.get(tag["Key"])
    ]
    tasks = []
    response = []

    if tags_to_remove := [
        tag["Key"] for tag in existing_tags if not template_tag_map.get(tag["Key"])
    ]:
        log_str = "Stale tags discovered."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.DETACH,
                resource_type="aws:identity_center:permission_set",
                resource_id=permission_set_arn,
                attribute="tags",
                change_summary={"TagKeys": tags_to_remove},
            )
        ]
        response.extend(proposed_changes)
        if ctx.execute:
            log_str = f"{log_str} Removing tags..."
            apply_awaitable = boto_crud_call(
                identity_center_client.untag_resource,
                InstanceArn=instance_arn,
                ResourceArn=permission_set_arn,
                TagKeys=tags_to_remove,
            )
            tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, tags=tags_to_remove, **log_params)

    if tags_to_apply:
        log_str = "New tags discovered in AWS."
        proposed_changes = [
            ProposedChange(
                change_type=ProposedChangeType.ATTACH,
                resource_type="aws:identity_center:permission_set",
                resource_id=permission_set_arn,
                attribute="tags",
                new_value=tag,
                current_value=existing_tag_map.get(tag["Key"]),
            )
            for tag in tags_to_apply
        ]
        response.extend(proposed_changes)

        if ctx.execute:
            log_str = f"{log_str} Adding tags..."
            apply_awaitable = boto_crud_call(
                identity_center_client.tag_resource,
                InstanceArn=instance_arn,
                ResourceArn=permission_set_arn,
                Tags=tags_to_apply,
            )
            tasks.append(plugin_apply_wrapper(apply_awaitable, proposed_changes))
        log.debug(log_str, tags=tags_to_apply, **log_params)

    if tasks:
        results: list[list[ProposedChange]] = await asyncio.gather(*tasks)
        return list(chain.from_iterable(results))
    else:
        return response


async def delete_permission_set(
    identity_center_client,
    instance_arn: str,
    permission_set_arn: str,
    current_permission_set: dict,
    account_assignments: list[dict],
    log_params: dict,
):
    tasks = []
    managed_policies = [
        mp["Arn"] for mp in current_permission_set.get("ManagedPolicies", [])
    ]
    customer_managed_policy_references = current_permission_set.get(
        "CustomerManagedPolicyReferences"
    )
    permissions_boundary = current_permission_set.get("PermissionsBoundary")
    inline_policy = current_permission_set.get("InlinePolicy")
    tags = current_permission_set.get("Tags")

    if account_assignments:
        log_params["account_assignments"] = account_assignments
        for assignment in account_assignments:
            tasks.append(
                delete_account_assignment(
                    identity_center_client,
                    account_id=assignment["account_id"],
                    instance_arn=instance_arn,
                    permission_set_arn=permission_set_arn,
                    resource_type=assignment["resource_type"],
                    resource_id=assignment["resource_id"],
                    resource_name=assignment["resource_name"],
                    log_params=log_params,
                )
            )

    if managed_policies:
        log_params["managed_policies"] = managed_policies
        tasks.extend(
            [
                boto_crud_call(
                    identity_center_client.detach_managed_policy_from_permission_set,
                    InstanceArn=instance_arn,
                    PermissionSetArn=permission_set_arn,
                    ManagedPolicyArn=policy_arn,
                )
                for policy_arn in managed_policies
            ]
        )
    if customer_managed_policy_references:
        log_params[
            "customer_managed_policy_references"
        ] = customer_managed_policy_references
        tasks.extend(
            [
                boto_crud_call(
                    identity_center_client.detach_customer_managed_policy_reference_from_permission_set,
                    retryable_errors=["ConflictException"],
                    InstanceArn=instance_arn,
                    PermissionSetArn=permission_set_arn,
                    CustomerManagedPolicyReference=policy,
                )
                for policy in customer_managed_policy_references
            ]
        )
    if permissions_boundary:
        log_params["permissions_boundary"] = permissions_boundary
        tasks.append(
            boto_crud_call(
                identity_center_client.delete_permissions_boundary_from_permission_set,
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
            )
        )
    if inline_policy:
        log_params["inline_policy"] = inline_policy
        tasks.append(
            boto_crud_call(
                identity_center_client.delete_inline_policy_from_permission_set,
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
            )
        )
    if tags:
        log_params["tags"] = tags
        tags_to_remove = [tag["Key"] for tag in tags]
        tasks.append(
            boto_crud_call(
                identity_center_client.untag_resource,
                InstanceArn=instance_arn,
                ResourceArn=permission_set_arn,
                TagKeys=tags_to_remove,
            )
        )

    log.debug(
        "Detaching resources from the permission set.",
        permission_set_arn=permission_set_arn,
        **log_params,
    )
    await async_batch_processor(tasks, 10, 1)

    retry_count = 0
    while True:
        try:
            await boto_crud_call(
                identity_center_client.delete_permission_set,
                InstanceArn=instance_arn,
                PermissionSetArn=permission_set_arn,
            )
        except ClientError as err:
            err_response = err.response["Error"]
            if (
                "PermissionSet has ApplicationProfile associated with it."
                in err_response["Message"]
            ) and retry_count < 10:
                # Wait for the detached resources to be deleted so that the permission set can be deleted.
                log.debug(
                    "Waiting for resources to be detached from the permission set.",
                    **log_params,
                )
                retry_count += 1
                await asyncio.sleep(5)
            elif err_response["Code"] == "ResourceNotFoundException":
                return
            else:
                raise



from __future__ import annotations

import itertools
import os
from collections import defaultdict
from typing import TYPE_CHECKING

import aiofiles

from iambic.core import noq_json as json
from iambic.core.logger import log
from iambic.core.models import ExecutionMessage
from iambic.core.template_generation import (
    create_or_update_template,
    delete_orphaned_templates,
)
from iambic.core.utils import NoqSemaphore, normalize_dict_keys, resource_file_upsert
from iambic.plugins.v0_1_0.aws.event_bridge.models import GroupMessageDetails
from iambic.plugins.v0_1_0.aws.iam.group.models import (
    AWS_IAM_GROUP_TEMPLATE_TYPE,
    AwsIamGroupTemplate,
    GroupProperties,
)
from iambic.plugins.v0_1_0.aws.iam.group.utils import (
    get_group_across_accounts,
    get_group_inline_policies,
    get_group_managed_policies,
    list_groups,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount
from iambic.plugins.v0_1_0.aws.template_generation import (
    base_group_str_attribute,
    group_dict_attribute,
    group_int_or_str_attribute,
)
from iambic.plugins.v0_1_0.aws.utils import (
    calculate_import_preference,
    get_aws_account_map,
)

if TYPE_CHECKING:
    from iambic.plugins.v0_1_0.aws.iambic_plugin import AWSConfig

RESOURCE_DIR = ["iam", "group"]


def get_response_dir(exe_message: ExecutionMessage, aws_account: AWSAccount) -> str:
    if exe_message.provider_id:
        return exe_message.get_directory(*RESOURCE_DIR)
    else:
        return exe_message.get_directory(aws_account.account_id, *RESOURCE_DIR)


def get_template_dir(base_dir: str) -> str:
    return str(os.path.join(base_dir, "resources", "aws", *RESOURCE_DIR))


def get_templated_group_file_path(
    group_dir: str,
    group_name: str,
    included_accounts: list[str],
) -> str:
    if included_accounts is not None and len(included_accounts) > 1:
        separator = "multi_account"
    elif included_accounts == ["*"] or included_accounts is None:
        separator = "all_accounts"
    else:
        separator = included_accounts[0]

    file_name = (
        group_name.replace(" ", "")
        .replace("{{var.", "")
        .replace("{{", "")
        .replace("}}_", "_")
        .replace("}}", "_")
        .replace(".", "_")
        .lower()
    )
    return str(os.path.join(group_dir, separator, f"{file_name}.yaml"))


async def generate_account_group_resource_files(
    exe_message: ExecutionMessage,
    aws_account: AWSAccount,
) -> dict:
    account_group_response_dir = get_response_dir(exe_message, aws_account)
    group_resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []

    response = dict(account_id=aws_account.account_id, groups=[])
    iam_client = await aws_account.get_boto3_client("iam")
    account_groups = await list_groups(iam_client)

    log.debug(
        "Retrieved AWS IAM Groups.",
        account_id=aws_account.account_id,
        account_name=aws_account.account_name,
        group_count=len(account_groups),
    )

    for account_group in account_groups:
        group_path = os.path.join(
            account_group_response_dir, f'{account_group["GroupName"]}.json'
        )
        response["groups"].append(
            {
                "path": group_path,
                "name": account_group["GroupName"],
                "account_id": aws_account.account_id,
            }
        )
        messages.append(
            dict(file_path=group_path, content_as_dict=account_group, replace_file=True)
        )

    await group_resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS IAM Groups.",
        account_id=aws_account.account_id,
        group_count=len(account_groups),
    )

    return response


async def generate_group_resource_file_for_all_accounts(
    exe_message: ExecutionMessage, aws_accounts: list[AWSAccount], group_name: str
) -> list:
    account_group_response_dir_map = {
        aws_account.account_id: get_response_dir(exe_message, aws_account)
        for aws_account in aws_accounts
    }
    group_resource_file_upsert_semaphore = NoqSemaphore(resource_file_upsert, 10)
    messages = []
    response = []

    group_across_accounts = await get_group_across_accounts(
        aws_accounts, group_name, False
    )
    group_across_accounts = {k: v for k, v in group_across_accounts.items() if v}

    log.debug(
        "Retrieved AWS IAM Group for all accounts.",
        group_name=group_name,
        total_accounts=len(group_across_accounts),
    )

    for account_id, account_group in group_across_accounts.items():
        group_path = os.path.join(
            account_group_response_dir_map[account_id],
            f'{account_group["GroupName"]}.json',
        )
        response.append(
            {
                "path": group_path,
                "name": account_group["GroupName"],
                "account_id": account_id,
            }
        )
        messages.append(
            dict(file_path=group_path, content_as_dict=account_group, replace_file=True)
        )

    await group_resource_file_upsert_semaphore.process(messages)
    log.debug(
        "Finished caching AWS IAM Group for all accounts.",
        group_name=group_name,
        total_accounts=len(group_across_accounts),
    )

    return response


async def set_group_resource_inline_policies(
    group_name: str, group_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    group_inline_policies = await get_group_inline_policies(group_name, iam_client)
    for k in group_inline_policies.keys():
        group_inline_policies[k]["policy_name"] = k

    group_inline_policies = list(group_inline_policies.values())
    await resource_file_upsert(
        group_resource_path, {"InlinePolicies": group_inline_policies}, False
    )


async def set_group_resource_managed_policies(
    group_name: str, group_resource_path: str, aws_account: AWSAccount
):
    iam_client = await aws_account.get_boto3_client("iam")
    group_managed_policies = await get_group_managed_policies(group_name, iam_client)
    await resource_file_upsert(
        group_resource_path, {"ManagedPolicies": group_managed_policies}, False
    )


async def _account_id_to_group_map(group_refs):
    account_id_to_group_map = {}
    for group_ref in group_refs:
        async with aiofiles.open(group_ref["path"], mode="r") as f:
            content_dict = json.loads(await f.read())

            account_id_to_group_map[group_ref["account_id"]] = normalize_dict_keys(
                content_dict
            )
    return account_id_to_group_map


async def create_templated_group(  # noqa: C901
    aws_account_map: dict[str, AWSAccount],
    group_name: str,
    group_refs: list[dict],
    group_dir: str,
    existing_template_map: dict,
    config: AWSConfig,
) -> AwsIamGroupTemplate:
    account_id_to_group_map = await _account_id_to_group_map(group_refs)
    num_of_accounts = len(group_refs)

    min_accounts_required_for_wildcard_included_accounts = (
        config.min_accounts_required_for_wildcard_included_accounts
    )

    # calculate preference based on existing template
    prefer_templatized = calculate_import_preference(
        existing_template_map.get(group_name)
    )

    # Generate the params used for attribute creation
    group_template_params = {"identifier": group_name}
    group_template_properties = {"group_name": group_name}
    path_resources = list()
    managed_policy_resources = list()
    inline_policy_document_resources = list()
    for account_id, group_dict in account_id_to_group_map.items():
        path_resources.append(
            {
                "account_id": account_id,
                "resources": [{"resource_val": group_dict["path"]}],
            }
        )

        if managed_policies := group_dict.get("managed_policies"):
            managed_policy_resources.append(
                {
                    "account_id": account_id,
                    "resources": [{"resource_val": mp} for mp in managed_policies],
                }
            )

        if inline_policies := group_dict.get("inline_policies"):
            # Normalize the inline policy statements to be a list
            for inline_policy in inline_policies:
                if isinstance(inline_policy.get("statement"), dict):
                    inline_policy["statement"] = [inline_policy["statement"]]
            inline_policy_document_resources.append(
                {
                    "account_id": account_id,
                    "resources": [
                        {"resource_val": inline_policy}
                        for inline_policy in inline_policies
                    ],
                }
            )

    if (
        len(group_refs) != len(aws_account_map)
        or len(aws_account_map) <= min_accounts_required_for_wildcard_included_accounts
    ):
        group_template_params["included_accounts"] = [
            aws_account_map[group_ref["account_id"]].account_name
            for group_ref in group_refs
        ]
    else:
        group_template_params["included_accounts"] = ["*"]

    path = await group_int_or_str_attribute(
        aws_account_map, num_of_accounts, path_resources, "path"
    )
    if path != "/":
        group_template_properties["path"] = path

    if managed_policy_resources:
        group_template_properties["managed_policies"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            managed_policy_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    if inline_policy_document_resources:
        group_template_properties["inline_policies"] = await group_dict_attribute(
            aws_account_map,
            num_of_accounts,
            inline_policy_document_resources,
            False,
            prefer_templatized=prefer_templatized,
        )

    file_path = get_templated_group_file_path(
        group_dir,
        group_name,
        group_template_params.get("included_accounts"),
    )
    return create_or_update_template(
        file_path,
        existing_template_map,
        group_name,
        AwsIamGroupTemplate,
        group_template_params,
        GroupProperties(**group_template_properties),
        list(aws_account_map.values()),
    )


async def collect_aws_groups(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    iam_template_map: dict,
    detect_messages: list[GroupMessageDetails] = None,
):
    aws_account_map = await get_aws_account_map(config)
    if exe_message.provider_id:
        aws_account_map = {
            exe_message.provider_id: aws_account_map[exe_message.provider_id]
        }

    if detect_messages:
        detect_messages = [
            msg for msg in detect_messages if isinstance(msg, GroupMessageDetails)
        ]
        if not detect_messages:
            return

    existing_template_map = iam_template_map.get(AWS_IAM_GROUP_TEMPLATE_TYPE, {})
    set_group_resource_inline_policies_semaphore = NoqSemaphore(
        set_group_resource_inline_policies, 20
    )
    set_group_resource_managed_policies_semaphore = NoqSemaphore(
        set_group_resource_managed_policies, 30
    )

    log.info(
        "Generating AWS group templates. Beginning to retrieve AWS IAM Groups.",
        accounts=list(aws_account_map.keys()),
    )

    if detect_messages:
        aws_accounts = list(aws_account_map.values())
        generate_group_resource_file_for_all_accounts_semaphore = NoqSemaphore(
            generate_group_resource_file_for_all_accounts, 45
        )
        tasks = [
            {"aws_accounts": aws_accounts, "group_name": group.group_name}
            for group in detect_messages
            if not group.delete
        ]

        # Remove deleted or mark templates for update
        deleted_groups = [group for group in detect_messages if group.delete]
        if deleted_groups:
            for group in deleted_groups:
                group_account = aws_account_map.get(group.account_id)
                if existing_template := existing_template_map.get(group.group_name):
                    if len(existing_template.included_accounts) == 1 and (
                        existing_template.included_accounts[0]
                        == group_account.account_name
                        or existing_template.included_accounts[0]
                        == group_account.account_id
                    ):
                        # It's the only account for the template so delete it
                        existing_template.delete()
                    else:
                        # There are other accounts for the template so re-eval the template
                        tasks.append(
                            {
                                "exe_message": exe_message,
                                "aws_accounts": aws_accounts,
                                "group_name": existing_template.properties.group_name,
                            }
                        )

        account_group_list = (
            await generate_group_resource_file_for_all_accounts_semaphore.process(tasks)
        )
        account_group_list = list(itertools.chain.from_iterable(account_group_list))
        account_group_map = defaultdict(list)
        for account_group in account_group_list:
            account_group_map[account_group["account_id"]].append(account_group)
        account_groups = [
            dict(account_id=account_id, groups=groups)
            for account_id, groups in account_group_map.items()
        ]
    elif exe_message.provider_id:
        aws_account = aws_account_map[exe_message.provider_id]
        account_groups = [
            (await generate_account_group_resource_files(exe_message, aws_account))
        ]
    else:
        generate_account_group_resource_files_semaphore = NoqSemaphore(
            generate_account_group_resource_files, 5
        )
        account_groups = await generate_account_group_resource_files_semaphore.process(
            [
                {"exe_message": exe_message, "aws_account": aws_account}
                for aws_account in aws_account_map.values()
            ]
        )

    if not any(account_group["groups"] for account_group in account_groups):
        log.info(
            "No groups found in any AWS accounts.",
            accounts=list(aws_account_map.keys()),
        )
        return

    messages = []
    # Upsert groups
    for account_group in account_groups:
        for group in account_group["groups"]:
            messages.append(
                {
                    "group_name": group["name"],
                    "group_resource_path": group["path"],
                    "aws_account": aws_account_map[account_group["account_id"]],
                }
            )

    log.info(
        "Setting inline policies in group templates",
        accounts=list(aws_account_map.keys()),
    )
    await set_group_resource_inline_policies_semaphore.process(messages)
    log.info(
        "Setting managed policies in group templates",
        accounts=list(aws_account_map.keys()),
    )
    await set_group_resource_managed_policies_semaphore.process(messages)
    log.info("Finished retrieving group details", accounts=list(aws_account_map.keys()))

    account_group_output = json.dumps(account_groups)
    with open(
        exe_message.get_file_path(*RESOURCE_DIR, file_name_and_extension="output.json"),
        "w",
    ) as f:
        f.write(account_group_output)


async def generate_aws_group_templates(
    exe_message: ExecutionMessage,
    config: AWSConfig,
    base_output_dir: str,
    iam_template_map: dict,
    detect_messages: list[GroupMessageDetails] = None,
):
    aws_account_map = await get_aws_account_map(config)
    existing_template_map = iam_template_map.get(AWS_IAM_GROUP_TEMPLATE_TYPE, {})
    group_dir = get_template_dir(base_output_dir)
    account_groups = await exe_message.get_sub_exe_files(
        *RESOURCE_DIR, file_name_and_extension="output.json", flatten_results=True
    )

    log.info("Grouping groups")
    # Move everything to required structure
    for account_group_elem in range(len(account_groups)):
        for group_elem in range(len(account_groups[account_group_elem]["groups"])):
            group_name = account_groups[account_group_elem]["groups"][group_elem].pop(
                "name"
            )
            account_groups[account_group_elem]["groups"][group_elem][
                "resource_val"
            ] = group_name

        account_groups[account_group_elem]["resources"] = account_groups[
            account_group_elem
        ].pop("groups", [])

    grouped_group_map = await base_group_str_attribute(aws_account_map, account_groups)

    log.debug("Writing templated groups")
    all_resource_ids = set()
    for group_name, group_refs in grouped_group_map.items():
        resource_template = await create_templated_group(
            aws_account_map,
            group_name,
            group_refs,
            group_dir,
            existing_template_map,
            config,
        )
        if not resource_template:
            # Template not updated. Most likely because it's an `enforced` template.
            continue
        all_resource_ids.add(resource_template.resource_id)

    if not detect_messages:
        # NEVER call this if messages are passed in because all_resource_ids will only contain those resources
        delete_orphaned_templates(
            list(existing_template_map.values()), all_resource_ids
        )

    log.info("Finished templated group generation")



from __future__ import annotations

import asyncio
from typing import Optional

from pydantic import BaseModel, Field, validator

from iambic.core.iambic_plugin import ProviderPlugin
from iambic.plugins.v0_1_0 import PLUGIN_VERSION
from iambic.plugins.v0_1_0.aws.handlers import (
    apply,
    aws_account_update_and_discovery,
    decode_aws_secret,
    detect_changes,
    import_aws_resources,
    load,
)
from iambic.plugins.v0_1_0.aws.iam.group.models import AwsIamGroupTemplate
from iambic.plugins.v0_1_0.aws.iam.policy.models import AwsIamManagedPolicyTemplate
from iambic.plugins.v0_1_0.aws.iam.role.models import AwsIamRoleTemplate
from iambic.plugins.v0_1_0.aws.iam.user.models import AwsIamUserTemplate
from iambic.plugins.v0_1_0.aws.identity_center.permission_set.models import (
    AwsIdentityCenterPermissionSetTemplate,
)
from iambic.plugins.v0_1_0.aws.models import AWSAccount, AWSOrganization


class AWSConfig(BaseModel):
    organizations: list[AWSOrganization] = Field(
        [], description="A list of AWS Organizations to be managed by iambic"
    )
    accounts: list[AWSAccount] = Field(
        [], description="A list of AWS Accounts to be managed by iambic"
    )
    min_accounts_required_for_wildcard_included_accounts: int = Field(
        3,
        description=(
            "Iambic will set included_accounts = * on imported resources "
            "that exist on all accounts if the minimum number of accounts is met."
        ),
    )
    sqs_cloudtrail_changes_queues: Optional[list[str]] = []
    spoke_role_is_read_only: bool = Field(
        False,
        description=(
            "aws iambic spoke role is configured as read_only. "
            "If true, it will restrict IAMbic capability in AWS"
        ),
    )

    @validator("organizations", allow_reuse=True)
    def validate_organizations(cls, organizations):
        if len(organizations) > 1:
            raise ValueError("Only one AWS Organization is supported at this time.")
        return organizations

    @validator("accounts", allow_reuse=True)
    def validate_unique_accounts(cls, accounts):
        account_ids = set()
        account_names = set()
        for account in accounts:
            if account.account_id in account_ids:
                raise ValueError(f"duplicate account_id found: {account.account_id}")
            account_ids.add(account.account_id)

            if account.account_name in account_names:
                raise ValueError(
                    f"duplicate account_name found: {account.account_name}"
                )
            account_names.add(account.account_name)
        return accounts

    @property
    def hub_role_arn(self):
        if self.organizations:
            return self.organizations[0].hub_role_arn
        else:
            return [
                account.hub_role_arn
                for account in self.accounts
                if account.hub_role_arn
            ][0]

    async def set_identity_center_details(self, account_id: str = None):
        if self.accounts:
            if account_id:
                account = next(
                    account
                    for account in self.accounts
                    if account.account_id == account_id
                )
                await account.set_identity_center_details()
            else:
                await asyncio.gather(
                    *[
                        account.set_identity_center_details()
                        for account in self.accounts
                    ]
                )

    async def get_boto_session_from_arn(self, arn: str, region_name: str = None):
        region_name = region_name or arn.split(":")[3]
        account_id = arn.split(":")[4]
        aws_account_map = {account.account_id: account for account in self.accounts}
        aws_account = aws_account_map[account_id]
        return await aws_account.get_boto3_session(region_name)


IAMBIC_PLUGIN = ProviderPlugin(
    config_name="aws",
    version=PLUGIN_VERSION,
    provider_config=AWSConfig,
    async_apply_callable=apply,
    async_import_callable=import_aws_resources,
    async_load_callable=load,
    async_decode_secret_callable=decode_aws_secret,
    async_detect_changes_callable=detect_changes,
    async_discover_upstream_config_changes_callable=aws_account_update_and_discovery,
    templates=[
        AwsIdentityCenterPermissionSetTemplate,
        AwsIamGroupTemplate,
        AwsIamRoleTemplate,
        AwsIamUserTemplate,
        AwsIamManagedPolicyTemplate,
    ],
)
